@@ -10,4 +10,11 @@
 #' @export
 #' @importFrom magrittr %>%
 #' @usage lhs \%>\% rhs
-NULL
\ No newline at end of file
+NULL
+
+#' @importFrom mosaicCore makeFun df_stats
+#' @export
+mosaicCore::makeFun
+
+#' @export
+mosaicCore::df_stats
\ No newline at end of file
